/* file.c -- Copyright (C)  2000-2008 Marco Cesati
 * Tue, 31 Aug 2004 15:21:16 +0200
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STRING_H
#  include <string.h>
#endif
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#if HAVE_SYS_TIME_H
#  include <sys/time.h>
#endif
#if HAVE_UNISTD_H
#  include <unistd.h>
#endif
#if HAVE_SYS_TIME_H
#  include <time.h>
#endif

#include "config.h"
#include "mktest.h"

#define BUFFERLEN 4096

#define NUMTOKEN    8

char *question[MAXQUESTIONS];
char *answer[MAXANSWERS];
int good[MAXANSWERS];
int nanswers[MAXQUESTIONS];
int point[MAXANSWERS];
int mulquestion[MAXQUESTIONS];

char *course = NULL;
char *advise = NULL;
char *teacher = NULL;
char *date = NULL;
char *preamble = NULL;
const char *fontsize = NULL;
int formversion = 0;
char maxanswer_letter = 0;
char *formdate = NULL;
char *formcourse = NULL;

#define panic(A...) do { fprintf(stderr,A); exit(EXIT_FAILURE); } while (0)

char const encode[] = {
    '$', /* -9 */
    '^', /* -8 */
    '/', /* -7 */
    '&', /* -6 */
    '%', /* -5 */
    '#', /* -4 */
    '+', /* -3 */
    '=', /* -2 */
    '-', /* -1 */
    '.', /*  0 */
    '1', /* +1 */        
    '2', /* +2 */        
    '3', /* +3 */        
    '4', /* +4 */        
    '5', /* +5 */        
    '6', /* +6 */        
    '7', /* +7 */        
    '8', /* +8 */        
    '9'  /* +9 */        
};

#define SHIFTENCODE (9)


void atd_header(int maxq)
{
    int i, j;
    struct timeval tv;
    
    if (!gettimeofday(&tv, NULL)) {
        printf("ATD ");
        for (i=-9; i<10; ++i)
            printf("%c", encode[SHIFTENCODE+i]);
        printf(" -- generated by %s -- %s", PACKAGE_STRING, ctime(&tv.tv_sec));
    } else {
        fprintf(stderr, "Error invoking gettimeofday()!\n");
        exit(EXIT_FAILURE);
    }
    
    printf("Form version: %d/%c\n", formversion, maxanswer_letter);

    for (i=0; i<=maxq; ++i) {
        printf("%d:", i+1);
        for (j=0; j<nanswers[i]; ++j)
            printf("%c", encode[SHIFTENCODE+
                        point[i*MAXANSWERSPERQUESTION+j]]);
        printf(" ");
    }
    putchar('\n');
}


static inline int get_token(void)
{
    int t = yylex();
    if (t == ERR)
        panic(_errors[_error], "ATQ", _linenum);
    return t;
}


int readfile(FILE *input)
{
    int tok;
    int cquestion = -1, canswer = 0;
    
    restart_atq_scanner(input);

    /* Paranoid */
    
    memset(question, 0, MAXQUESTIONS*sizeof(char *));
    memset(answer, 0, MAXANSWERS*sizeof(char *));
    memset(good, 0, MAXANSWERS*sizeof(int));
    memset(nanswers, 0, MAXQUESTIONS*sizeof(int));
    memset(point, 0, MAXANSWERS*sizeof(int));
    memset(mulquestion, 0, MAXQUESTIONS*sizeof(int));
      
    tok = get_token();  
    
    while (tok) {
    
        switch (tok) {
                            
            case QUESTION:
                if (cquestion > -1) {
                    nanswers[cquestion] = canswer;
                    if (maxanswer_letter < canswer+'A'-1)
                        maxanswer_letter = canswer+'A'-1;
                }
                ++cquestion;
                
                if (get_token() != ARG)
                    panic("'{' expected at line %d\n", _linenum);
                                    
                question[cquestion] = strdup(yytext);
                canswer = 0;
            break;

            case MULQUESTION:
                if (cquestion > -1) {
                    nanswers[cquestion] = canswer;
                    if (maxanswer_letter < canswer+'A'-1)
                        maxanswer_letter = canswer+'A'-1;
                }
                ++cquestion;
                if (get_token() != ARG)
                    panic("'{' expected at line %d\n", _linenum);
                question[cquestion] = strdup(yytext);
                mulquestion[cquestion]++;
                canswer = 0;
            break;
                    
            case GOOD:
                if (get_token() != ARG)
                    panic("'{' expected at line %d\n", _linenum); 
                answer[cquestion*MAXANSWERSPERQUESTION+canswer] = strdup(yytext);
                good[cquestion*MAXANSWERSPERQUESTION+canswer] = 1;
                if (get_token() != NUM)
                    panic("Number expected at line %d\n", _linenum);
                point[cquestion*MAXANSWERSPERQUESTION+canswer] = _value;
                ++canswer;
            break;
                    
            case BAD:
                if (get_token() != ARG)
                    panic("'{' expected at line %d\n", _linenum); 
                answer[cquestion*MAXANSWERSPERQUESTION+canswer] = strdup(yytext);
                if (get_token() != NUM)
                    panic("Number expected at line %d\n", _linenum);
                point[cquestion*MAXANSWERSPERQUESTION+canswer] = -_value;
                ++canswer;
            break;
                    
            case COURSE:
                if (get_token() != ARG)
                    panic("'{' expected at line %d\n", _linenum); 
                course = strdup(yytext);
            break;
                    
            case TEACHER:
                if (get_token() != ARG)
                    panic("'{' expected at line %d\n", _linenum); 
                teacher = strdup(yytext);
            break;
                    
            case DATE:
                if (get_token() != ARG)
                    panic("'{' expected at line %d\n", _linenum); 
                date = strdup(yytext);
            break;
                                        
            case ADVISE:
                if (get_token() != ARG)
                    panic("'{' expected at line %d\n", _linenum); 
                advise = strdup(yytext);           
            break;
                    
            case GROUPS:
                /* eat integer arguments */
                 while ((tok = get_token()) == NUM);
                continue;
            break;

            case FONTPITCH:
                 if (get_token() != NUM)
                     panic("Number expected after \\fontpitch{, line %d\n",
                           _linenum);
                 if (_value < 10 || _value > 12)
                     panic("Invalid argument (%d) in \\fontpitch, line %d\n",
                           _value, _linenum);
                 fontsize = fontsize_str(_value);
            break;

            case ATQFILE:
                 panic("Error: mktest cannot handle the \\atqfile command!\n"
                       "(Perhaps an \\atqfile command is included in an ATQ sub-file,\n"
                       "or mktest was invoked directly on a file containing an \\atqfile command)\n");
            break;

            case PREAMBLE:
                 if (get_token() != ARG)
                     panic("'{' expected after \\preamble at line %d\n", _linenum);
                 preamble = strdup(yytext);
            break;

            case FORMVERSION:
                if (get_token() != NUM)
                    panic("Number expected after \\formversion{, line %d\n",
                          _linenum);
                    formversion = _value;
            break;

            case FORMDATE:
                if (get_token() != ARG)
                    panic("'{' expected after \\formdate at line %d\n", _linenum);
                formdate = strdup(yytext);
            break;

            case FORMCOURSE:
                if (get_token() != ARG)
                    panic("'{' expected after \\formcourse at line %d\n", _linenum);
                formcourse = strdup(yytext);
            break;

            default:
                fprintf(stderr, "Unexpected token (%d), exiting...\n",
                        tok);
                exit(EXIT_FAILURE);
            break;
        } /* switch */
        tok = get_token();
    } /* while */
    nanswers[cquestion] = canswer;
    if (maxanswer_letter < canswer+'A'-1)
        maxanswer_letter = canswer+'A'-1;
    return cquestion;
}
