/* main.c -- Copyright (C)  2000-2008 Marco Cesati
 * Fri, 18 Feb 2005 13:13:50 +0100
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif

#include "rdform.h"
#include "form.h"

#define VERSION_AND_COPYRIGHT \
          PACKAGE " version " PACKAGE_VERSION \
          " -- Copyright (C) 2000-2008 Marco Cesati <cesati@uniroma2.it>\n" \
          "This is free software; see the source for copying conditions.\n" \
          "There is NO warranty; not even for MERCHANTABILITY or FITNESS " \
          "FOR A PARTICULAR PURPOSE.\n\n"


int quiet = 0;
int verbose = 0;

static int get_form_version(char *arg)
{
    int l;
    char *p;
    
    l = (int) strtol(arg, &p, 10);
    if (*p || l < 0 || l > max_form_version())
        l = -1;
    return l; 
}

static inline void abort_usage(char *progname)
{
    fprintf(stderr, VERSION_AND_COPYRIGHT);
    fprintf(stderr, "Usage: %s [-t] [-p] [-q] [-v] [-f <num>] <image file>\n",
            progname);
    fprintf(stderr, "Options:\n\t-t\t\tConvert image in text on stdout\n"
	    "\t-q\t\tQuiet mode\n"
        "\t-v\t\tVerbose mode\n"
        "\t-f <num>\tSelect form version number (0-%d)\n"
        "\t-p\t\tImage file format is PPM (Portable Pixmap)\n"
	    "\t\t\tdefault file format is PBM (Portable Bitmap)\n"
        , max_form_version());
    exit(EXIT_FAILURE);
}

int main(int argc, char *argv[])
{    
    int i, code;
    int fnum = 0; /* default form has number 0 */
    int output_text = 0;
    int pbm_format = 1;
    struct image_t image;
    char *quiz;
        
    for (i=1; i<argc; ++i) {
        if (argv[i][0] == '-' && argv[i][1] == 't' && argv[i][2] == '\0')
            ++output_text;
        else if (argv[i][0] == '-' && argv[i][1] == 'p' && argv[i][2] == '\0')
            pbm_format = 0;
        else if (argv[i][0] == '-' && argv[i][1] == 'q' && argv[i][2] == '\0')
            ++quiet;
        else if (argv[i][0] == '-' && argv[i][1] == 'v' && argv[i][2] == '\0')
            ++verbose;
        else if (argv[i][0] == '-' && argv[i][1] == 'f') {
            if (argv[i][2] == '\0')
                fnum = get_form_version(argv[++i]);
            else
                fnum = get_form_version(argv[i]+2);
        } else
            break;  
    }        
   
    if (argc < 2 || i != argc-1 || fnum == -1)
        abort_usage(argv[0]);
   
    if (!quiet)
        fprintf(stderr, VERSION_AND_COPYRIGHT);
     
    set_form_version(fnum);
    if (!quiet)
        fprintf(stderr, 
                "Form version #%d: %d questions, %d answers/q\n", 
                fnum, form->answer.nquest, form->answer.nanswer);
    
    quiz = (char *) malloc(sizeof(char)*form->answer.nquest);
    if (!quiz) {
        fprintf(stderr, "Memory allocation error\n");
        return EXIT_FAILURE;
    }
    
    if (pbm_format)
        read_pbm_file(argv[i], &image);
    else
        read_ppm_file(argv[i], &image);
    
    if (output_text) {
        dump_to_text(image);
        return EXIT_SUCCESS;
    }

    compute_resolution(&image);
    
    code = extract_form_code(image);
    
    if (code == -2) {
        fprintf(stderr, "Invalid form code.\n");
        return EXIT_FAILURE;
    }
    
    if (code == -1)
        printf("---: ");        
    else
        printf("%03d: ", code);
        
    extract_marks(quiz, image);

    for (i=0; i<form->answer.nquest; i++)
        printf("[%s] ", str_answers(quiz[i]));
        
    printf("\n");
    return EXIT_SUCCESS;
}
