/* score.c -- Copyright (C)  2000-2008 Marco Cesati
 * Wed, 25 Feb 2004 16:28:04 +0100
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#if HAVE_STRING_H
#  include <string.h>
#endif

#define VERSION_AND_COPYRIGHT \
          PACKAGE " version " PACKAGE_VERSION \
          " -- Copyright (C) 2000-2008 Marco Cesati <cesati@uniroma2.it>\n\n" \
          "This is free software; see the source for copying conditions.\n" \
          "There is NO warranty; not even for MERCHANTABILITY or FITNESS " \
          "FOR A PARTICULAR PURPOSE.\n\n"


#define RDFORMOPTIONS "-q -f"
#define RDFORMOFFSET 12
    
char const extension[] = ".atd";

const char * rdformdefault = "rdform";

int formversion;
char maxanswer_letter;

#define MAXQUESTIONS           1000
#define MAXQUIZ                  20
#define MAXANSWERSPERQUESTION    26 
#define MAXANSWERS      (MAXQUESTIONS*MAXANSWERSPERQUESTION)
#define MAXFORMS               1000
#define BUFFERLEN 8192

int tab_mode = 0;
int neg_mode = 0;
int values[MAXQUESTIONS][MAXANSWERSPERQUESTION];
int point[MAXFORMS][MAXQUIZ][MAXANSWERSPERQUESTION];
int mulquestion[MAXFORMS][MAXQUIZ];

#define SHIFTENCODE (9)

char const encode[] = {
    '$', /* -9 */
    '^', /* -8 */
    '/', /* -7 */
    '&', /* -6 */
    '%', /* -5 */
    '#', /* -4 */
    '+', /* -3 */
    '=', /* -2 */
    '-', /* -1 */
    '.', /*  0 */
    '1', /* +1 */        
    '2', /* +2 */        
    '3', /* +3 */        
    '4', /* +4 */        
    '5', /* +5 */        
    '6', /* +6 */        
    '7', /* +7 */        
    '8', /* +8 */        
    '9'  /* +9 */        
};


void error_abort(const char * mesg)
{
    fprintf(stderr, "%s\n", mesg);
    exit(EXIT_FAILURE);
}

inline void skip_blanks(char **p, FILE *f)
{
    do {
        while (**p && (**p == ' ' || **p == '\t'))
            ++*p;
        if ((!**p) && (!fgets(*p, BUFFERLEN, f)))
            return;
    } while (**p == ' ' || **p == '\t');
}

void evaluate_results(FILE *f, FILE *out)
{
    char buffer[BUFFERLEN];    
    while (!feof(f)) {
        if (fgets(buffer, BUFFERLEN, f)) {
            char *p = buffer;
            int form = 0;
            int quiz = 0;
            int form_value = 0;
                                    
            skip_blanks(&p, f);
            if (!*p || *p == '\n')
                continue;
                
            if (*p < '0' || *p > '9') {
                if (*p++ == '-' && *p++ == '-' && *p == '-')
                    fprintf(stderr, "\nForm number is unreadable\n");
                else
                    fprintf(stderr, 
                        "\nInvalid format: 'digit' expected (1)\n");           
                return;
            }
            
            while (*p && *p >= '0' && *p <= '9') {
                form = form * 10 + (*p) - '0';
                ++p;
            }
                            
            if (!*p || *p++ != ':')
                error_abort("\nInvalid format: ':' expected");
            
            if (!*p || *p++ != ' ') 
                error_abort("\nInvalid format: ' ' expected");

            if (tab_mode)           
                fprintf(out, "%03d:\t", form);
            else
                fprintf(out, "Form %03d: ", form);
            
            do {
                int question_value = 0;
                skip_blanks(&p, f);
                if (*p && *p != '\n') {
                    if (*p++ != '[') 
                        error_abort("\nInvalid format: '[' expected");
                    if (!mulquestion[form][quiz]) {
                        if (*p && *p >= 'A' && *p <= maxanswer_letter) {
                            question_value = point[form][quiz][*p-'A'];
                            p++;
                            while (*p && *p >= 'A' && *p <= maxanswer_letter) {
                                question_value = 0;
                                p++;
                            }
                        }                        
                    } else {
                        if (neg_mode) {
#if 0
                            int _fl[5] = { 0, 0, 0, 0, 0 };
                            int _i;

                            while (*p && *p >= 'A' && *p <= maxanswer_letter) {
                                _fl[*p-'A']++;
                                p++;
                            }
                            for (_i=0; _i<5; _i++)
                                if (_fl[_i])
                                    question_value += point[form][quiz][_i];
                                else
                                    question_value -= point[form][quiz][_i];
#else
                            char _c = 'A';
                            while (*p && *p >= 'A' && *p <= maxanswer_letter) {
                                question_value += point[form][quiz][*p-'A'];
                                for (; _c < *p; _c++)
                                    question_value -= point[form][quiz][_c-'A'];
                                ++_c;
                                ++p;
                            }
                            for (; _c <= maxanswer_letter; _c++)
                                question_value -= point[form][quiz][_c-'A'];
#endif
                        } else {
                           while (*p && *p >= 'A' && *p <= maxanswer_letter) {
                                question_value += point[form][quiz][*p-'A'];
                                p++;
                            }
                        }
                    }
                    
                    if (!*p || *p !=']')
                        error_abort("\nInvalid format: ']' expected");
                    if (tab_mode)
                        fprintf(out, "%d\t", question_value);
                    else
                        fprintf(out, "%2d ", question_value);
                    form_value += question_value;
                    ++quiz;
                    p++;
                }
            } while (*p && *p != '\n');
            if (tab_mode)
                fprintf(out, "%d\n", form_value);
            else
                fprintf(out, ": %2d\n", form_value);
        }  /* if */
    } /* while */
}



void read_atd_file(FILE *f)
{
    char buffer[BUFFERLEN];
    int tf = 0;
    int maxform = -1;
    int compatibility_hack = 0;

    if (!fgets(buffer, BUFFERLEN, f)) 
        error_abort("Error reading ATD file (empty)");
    
    if (strncmp(buffer, "ATD ", 4)) {
        rewind(f);
        compatibility_hack++;
    } else {
        char *p = buffer+4;
        int i;
        for(i=0; i<19; ++i, ++p)
            if (*p != encode[i])
                error_abort("Unknown encoding format");
        if (!fgets(buffer, BUFFERLEN, f))
            error_abort("Error in ATD file (truncated)");

        if (sscanf(buffer, "Form version: %d/%c", &formversion,
                 &maxanswer_letter) != 2 || 
                 maxanswer_letter <= 'A' || maxanswer_letter > 'Z')
            error_abort("Error in ATD file (invalid form version)");

        if (!fgets(buffer, BUFFERLEN, f))
            error_abort("Error in ATD file (truncated)");

        for (i=0, p = buffer; *p && *p != '\n'; ++p, ++i) {
            int n = 0, a = 0;
            for (; *p && (*p <= '9' && *p >= '0'); ++p)
                n = n * 10 + *p - '0';
            if (*p != ':' && *p != '\n') 
                error_abort("Error in ATD file (no separator)");
            if (i != n-1)
                error_abort("Error in ATD file (unordered questions)");
            for (p++; *p && (*p != ' ' && *p != '\n'); ++p, ++a) {
                int j;
                for (j=0; j<19; ++j)
                    if (encode[j] == *p)
                        break;
                if (j==19)
                    error_abort("Error in ATD file (illegal encoding)");
                values[n-1][a] = j - SHIFTENCODE;
            }
        }            
    }
            
    while (!feof(f)) {
        if (fgets(buffer, BUFFERLEN, f)) {
            char *p = buffer;
            int form = 0;
            int quiz = 0;
                        
            skip_blanks(&p, f);
            if (!*p || *p == '\n')
                continue;
                
            if (*p < '0' || *p > '9')
                error_abort("\nInvalid format in ATD file: "
                        "'digit' expected (1)");
            
            while (*p && *p >= '0' && *p <= '9') {
                form = form * 10 + (*p) - '0';
                ++p;
            }
            
            if (form > maxform)
                maxform = form;
                
            if (!*p || *p++ != ':')
                error_abort("\nInvalid format in ATD file: "
                    "':' expected");

            do {
                int tot = 0;
                int ans = 0;
                
                skip_blanks(&p, f);
                if (!*p || *p == '\n')
                    break;
                    
                if (*p < '0' || *p > '9') 
                    error_abort("\nInvalid format in ATD file; "
                        "'digit' expected (2)");
                
                while (*p && *p >= '0' && *p <= '9') {
                    tot = tot * 10 + (*p) - '0';
                    ++p;
                }
                
                
                if (!*p || (*p != '(' && *p != '['))
                    error_abort("\nInvalid format in ATD file: "
                        "'(' or '[' expected");
                
                if (*p++ == '[') {
                    mulquestion[form][quiz]++;
                }
                while (*p && (*p != ')' && *p != ']')) {
                    if (*p < 'A' || *p > maxanswer_letter)
                        error_abort("\nInvalid format in ATD file: "
                               "answer code not valid");
                    if (compatibility_hack) {
                        int i;
                        for (i=0; i<1+(SHIFTENCODE*2); i++) {
                            if (encode[i] == *p) {
                                point[form][quiz][ans] = i-SHIFTENCODE;
                                break;
                            }
                        }
                        if (i == 1+(SHIFTENCODE*2)) 
                            error_abort("\nInvalid format in ATD file: "
                                "unknown encoding for score value");
                    } else 
                        point[form][quiz][ans] = values[tot-1][*p-'A'];   
                    ++p;
                    ++ans;
                }
                if (!*p)
                    error_abort("\nInvalid format in ATD file: "
                        "')' or ']' expected");
                ++quiz;
                p++;
            } while (*p);
        } /* if */
        ++tf;
    } /* while */
#if 0
{
    int i, j, k;
    for (i=0; i<=maxform; ++i) {
        fprintf(stderr, "Form %03d:", i);
        for (j=0; j<MAXQUIZ; ++j) {
            fprintf(stderr, " [");
            for (k=0; k<MAXANSWERSPERQUESTION; ++k)
                fprintf(stderr, " %d", point[i][j][k]);
            fprintf(stderr, " ]");
        }
        fprintf(stderr, "\n");
    }                
}   
#endif
#if 0
{
    int i, j;
    for (i=0; i<=maxform; ++i) {
        fprintf(stderr, "%03d: ", i);
        for (j=0; j<MAXQUIZ; ++j) {
            fprintf(stderr, " %d", mulquestion[i][j]);
        }
        fprintf(stderr, "\n");
    }
        
}
#endif
}

inline void usage(const char * prog)
{
        fprintf(stderr, VERSION_AND_COPYRIGHT);
        fprintf(stderr, "Usage: %s [-t] [-n] [--] <atd file> "
                         "[<pnm file> ...]\n", prog);
        exit(EXIT_FAILURE);
}

int main(int argc, char *argv[])
{
    FILE *atdf;
    char *filename;
    char *basename;
    int l;
    int na = 1;

        

    /* Pure paranoid */
        
    /* Static initialization appears to enlarge the code executable.
        No fs holes?! :-( */

    memset(values, 0, MAXQUESTIONS*MAXANSWERSPERQUESTION*sizeof(int));
    memset(point, 0, MAXFORMS*MAXQUIZ*MAXANSWERSPERQUESTION*sizeof(int));
    memset(mulquestion, 0, MAXFORMS*MAXQUIZ*sizeof(int));    
            
    if (argc < 2)
        usage(argv[0]);
    
    while (argv[na] && argv[na][0] == '-' && argv[na][2] == '\0') {
       int c = argv[na][1];
       if (c == 't')
            ++tab_mode;
        else if (c == 'n')
            ++neg_mode;
        else if (c == '-') {
            ++na;
            break;
        } else 
            usage(argv[0]);
        ++na;
    }
    
    if (!argv[na] || !argv[na][0])
        usage(argv[0]);
    
    filename = argv[na];
    l = strlen(filename);
    
    if (filename[l-4] != extension[0] ||
        filename[l-3] != extension[1] ||
        filename[l-2] != extension[2] ||
        filename[l-1] != extension[3]) {
    
        filename = (char *) malloc(sizeof(char)*(strlen(argv[na])+5));
        if (!filename) {
            fprintf(stderr, "Memory allocation error\n");
            return EXIT_FAILURE;
        }
        
        strcpy(filename, argv[na]);
        
        filename[l] = extension[0];
        filename[l+1] = extension[1];
        filename[l+2] = extension[2];
        filename[l+3] = extension[3];
        filename[l+4] = '\0';
        basename = argv[na];
    } else {
        basename = (char *) malloc(sizeof(char)*(strlen(argv[na])+1));
        if (!basename) {
            fprintf(stderr, "Memory allocation error\n");
            return EXIT_FAILURE;
        }
        strcpy(basename, argv[na]);
        basename[strlen(basename)-4] = '\0';
    }
    atdf = fopen(filename, "r");
    if (!atdf) {
        perror(filename);
        return EXIT_FAILURE;
    }    
           
    read_atd_file(atdf);

    if (fclose(atdf))
        perror(filename);

    if (argc > na+1) {
        char *ev;
        char buffer[BUFFERLEN];
        FILE *pf;
        FILE *of;

        sprintf(buffer, "%s.atr", basename);
        of = fopen(buffer, "w");
        if (!of) {
            perror(buffer);
            return EXIT_FAILURE;
        }
        
        if (tab_mode)
            fprintf(of, "FID:\tpunt1\tpunt2\tpunt3\tpunt4\tpunt5\tpunt6"
                        "\tpunt7\tpunt8\tpunt9\tpunt10\ttotale\n");
                    
        ev = getenv("RDFORMCMD");
 
        for (l=na+1; l<argc; ++l) {
            fprintf(stderr, "[%d: %s", l-na, argv[l]);
            if (!ev || !*ev)
                ev = (char *) rdformdefault;
            sprintf(buffer, "%s %s%d %s", ev, RDFORMOPTIONS, formversion, argv[l]);
            pf = popen(buffer, "r");
            evaluate_results(pf, of);
            pclose(pf); /* FIXME check error code */
            fprintf(stderr, "] ");
        }
        fprintf(stderr, "\n");
        if (fclose(of))
            fprintf(stderr, "Error closing output file\n");
 
    } else
        evaluate_results(stdin, stdout);
    
    return EXIT_SUCCESS;
}
